//---------------------------------------------------------------------------
// AnisotropicDisperse.cpp
//---------------------------------------------------------------------------
#include "IsotropicDisperse.h"
#include "SimManager.h"
#include "Plot.h"
#include "ModelMath.h"
#include <math.h>
#include <stdlib.h>

///////////////////////////////////////////////////////////////////////////////
//     Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
clIsotropicDisperse::clIsotropicDisperse( clSimManager * p_oSimManager ) : 
     clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager ),
     clDisperseBase( p_oSimManager), clGridBasedSpatialDisperse( p_oSimManager )
     {
       try
       {
         strcpy( m_cNameString, "isotropic disperse" );
         mp_fProb = NULL;
       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clIsotropicDisperse::clIsotropicDisperse" );
         throw( stcErr );
       }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
clIsotropicDisperse::~clIsotropicDisperse()
{
  short int i; //loop counter

  if ( mp_fProb )
  {
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {
      delete[] mp_fProb[i];
    }
  }

  delete[] mp_fProb;
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   CalculateNormalizer
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
float clIsotropicDisperse::CalculateNormalizer(int iSpecies )
{
  try
  {
    float fNormalizer = 0, //normalizer to be returned
          fIncr = 0.25, //distance increment
          fExp, //value of the exponential for a distance
          fDist; //distance counter
    int iMaxDist = 500;
    short int iSpIndex; //species index

    iSpIndex = mp_iIndexes[iSpecies];

    //What is the function?
    if ( lognormal == mp_iWhatFunction[iSpIndex] )
    {

      ///////////////////////////////////////
      //Lognormal
      ///////////////////////////////////////

      //The normalizer increments in units of 0.25 m2 (50 x 50 cm)

      //Calculate the exponential at a distance midway between the current
      //loop distance and the one behind it
      for ( fDist = fIncr; fDist < iMaxDist; fDist += fIncr )
      {
        fExp = clModelMath::CalculateLognormalFunction( mp_fDispersalX0[iSpIndex], mp_fThetaXb[iSpIndex], fDist - ( fIncr / 2 ) );
        fExp *= M_PI * ( pow( fDist, 2 ) - pow( fDist - fIncr, 2 ) );
        //this formulation stays in units of m2
        fNormalizer += fExp;
      }
    }
    else if ( weibull == mp_iWhatFunction[iSpIndex] )
    {
      ///////////////////////////////////////
      //Weibull isotropic
      ///////////////////////////////////////

      //The normalizer increments in units of 0.25 m2 (50 x 50 cm)

      //Calculate the exponential at a distance midway between the current
      //loop distance and the one behind it
      for ( fDist = fIncr; fDist < iMaxDist; fDist += fIncr )
      {
        fExp = clModelMath::CalculateWeibullFunction( mp_fDispersalX0[iSpIndex], mp_fThetaXb[iSpIndex], fDist - ( fIncr / 2 ) );
        fExp *= M_PI * ( pow( fDist, 2 ) - pow( fDist - fIncr, 2 ) );
        //this formulation stays in units of m2
        fNormalizer += fExp;
      }
    }

    //Since we incremented in units of 0.25 m2, multiply total by 0.25 to give a
    //normalizer for units of m2
 //   fNormalizer *= fIncr;
    return fNormalizer;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clIsotropicDisperse::CalculateNormalizer" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   CalculateNormalizer
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
/*float clIsotropicDisperse::CalculateNormalizer2( int iMaxDist, int iSpecies )
{
  try
  {
    double //total, x, y, mean, angle,
         fNormalizer = 0, //normalizer to be returneD
         fExp; //value of the exponential for a distance
    float fDist; //distance counter
    short int iSpIndex; //species index

    iSpIndex = mp_iIndexes[iSpecies];

    //What is the function?
    if ( lognormal == mp_iWhatFunction[iSpIndex] )
    {

      ///////////////////////////////////////
      //Lognormal
      ///////////////////////////////////////

      //The normalizer increments in units of 0.25 m2 (50 x 50 cm)

      //Calculate the exponential at a distance midway between the current
      //loop distance and the one behind it
      for ( fDist = 1; fDist <= iMaxDist; fDist++ )
      {
        fExp = CalculateLognormalFunction( mp_fDispersalX0[iSpIndex], mp_fThetaXb[iSpIndex], fDist - 0.5 );
        //this formulation stays in units of m2
        fNormalizer += fExp;
      }
    }
    else if ( weibull == mp_iWhatFunction[iSpIndex] )
    {
      ///////////////////////////////////////
      //Weibull isotropic
      ///////////////////////////////////////

      //The normalizer increments in units of 0.25 m2 (50 x 50 cm)

      //Calculate the exponential at a distance midway between the current
      //loop distance and the one behind it
      for ( fDist = 1; fDist <= iMaxDist; fDist++ )
      {
        fExp = CalculateWeibullFunction( mp_fDispersalX0[iSpIndex], mp_fThetaXb[iSpIndex], fDist - 0.5 );
        //this formulation stays in units of m2
        fNormalizer += fExp;
      }
    }

    return fNormalizer;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clIsotropicDisperse::CalculateNormalizer" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------
*/

///////////////////////////////////////////////////////////////////////////////
//   CalculateProbs
///////////////////////////////////////////////////////////////////////////////
void clIsotropicDisperse::CalculateProbs()
{
  try
  {
    float fNormalizer;
    int i, j, iSpIndex;

    //Determine what the max search radius as an integer is
    m_iMaxSearchRadius = (int)ceil( m_fMaxSearchRadius );

    //Declare the probability arrays
    mp_fProb = new float * [m_iNumBehaviorSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {
      mp_fProb[i] = new float[m_iMaxSearchRadius];
    }

    //Now create the values in the arrays for each species
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {

      iSpIndex = mp_iIndexes[mp_iWhatSpecies[i]];

      //Calculate the normalizer
      fNormalizer = CalculateNormalizer( mp_iWhatSpecies[i] );

      //Since we're dividing by it, take its inverse so we can multiply instead
      fNormalizer = 1.0 / fNormalizer;

      //Loop out the array and calculate the function distance divided by the
      //normalizer
      for ( j = 0; j < m_iMaxSearchRadius; j++ )
      {
        if ( lognormal == mp_iWhatFunction[iSpIndex] )
        {
          //Normally there would be an area term - but I want an area of one
          //square meter so I'm leaving out that term
          mp_fProb[i] [j] = fNormalizer
               * clModelMath::CalculateLognormalFunction( mp_fDispersalX0[iSpIndex], mp_fThetaXb[iSpIndex], j+1 );
        }
        else
        {
          //Normally there would be an area term - but I want an area of one
          //square meter so I'm leaving out that term
          mp_fProb[i] [j] = fNormalizer
               * clModelMath::CalculateWeibullFunction( mp_fDispersalX0[iSpIndex], mp_fThetaXb[iSpIndex], j+1 );
        }
      }
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clIsotropicDisperse::CalculateProbs" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   DoChildSetup
///////////////////////////////////////////////////////////////////////////////
void clIsotropicDisperse::DoChildSetup( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    CalculateProbs();
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clIsotropicDisperse::DoSetup" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   GetFunctionValue
///////////////////////////////////////////////////////////////////////////////
/*float clIsotropicDisperse::GetFunctionValue( float & fCellX, float & fCellY, float & fTreeX, float & fTreeY,
   int & iSpecies )
     {
       try
       {

         //Get the distance
         int iDistance = RandomRound( mp_oPlot->GetDistance( fCellX, fCellY, fTreeX, fTreeY ) );
         if ( iDistance < m_iMaxSearchRadius )
         {
           return mp_fProb[mp_iIndexes[iSpecies]] [iDistance];
         }
         else
         {
           return 0.0;
         }
       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clIsotropicDisperse::GetFunctionValue" );
         throw( stcErr );
       }
} */
//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   GetFunctionValue
///////////////////////////////////////////////////////////////////////////////
float clIsotropicDisperse::GetFunctionValue(float & fCellX, float & fCellY, float & fTreeX, float & fTreeY,
   int & iSpecies )
     {
       try
       {

         //Get the distance
         float fDistance = clModelMath::RandomRound( mp_oPlot->GetDistance( fCellX, fCellY, fTreeX, fTreeY ) );

         if ( fDistance < m_fMaxSearchRadius )
         {
           if ( lognormal == mp_iWhatFunction[mp_iIndexes[iSpecies]] )
        {
          //Normally there would be an area term - but I want an area of one
          //square meter so I'm leaving out that term
          return  M_PI * (2 * fDistance) * clModelMath::CalculateLognormalFunction( mp_fDispersalX0[mp_iIndexes[iSpecies]], mp_fThetaXb[mp_iIndexes[iSpecies]], fDistance )/CalculateNormalizer(iSpecies);
        }
        else
        {
          //Normally there would be an area term - but I want an area of one
          //square meter so I'm leaving out that term
          return clModelMath::CalculateWeibullFunction( mp_fDispersalX0[mp_iIndexes[iSpecies]], mp_fThetaXb[mp_iIndexes[iSpecies]], fDistance )/CalculateNormalizer(iSpecies);
        }
         }
         else
         {
           return 0.0;
         }
       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clIsotropicDisperse::GetFunctionValue" );
         throw( stcErr );
       }
}
//---------------------------------------------------------------------------

