//---------------------------------------------------------------------------
// MerchantableStemWoodVolume.cpp
//---------------------------------------------------------------------------
#include "MerchantableStemWoodVolume.h"
#include "TreePopulation.h"
#include "SimManager.h"
#include "ParsingFunctions.h"
#include "Constants.h"
#include <math.h>

////////////////////////////////////////////////////////////////////////////
// Constructor()
////////////////////////////////////////////////////////////////////////////
clMerchantableStemWoodVolume::clMerchantableStemWoodVolume( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
{
  try
  {
    //Set the namestring
    strcpy( m_cNameString, "FIA Merchantable Stem Wood Volume" );

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;

    //Indicate that this behavior intends to add one tree float data member
    m_iNewTreeFloats = 1;

    mp_fB1 = NULL;
    mp_fB2 = NULL;
    mp_fSpgr = NULL;
    mp_iIndexes = NULL;
    m_cQuery = NULL;
    mp_iVolumeCodes = NULL;
    mp_iBiomassCodes = NULL;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clMerchantableStemWoodVolume::clMerchantableStemWoodVolume" );
    throw( stcErr );
  }
}

////////////////////////////////////////////////////////////////////////////
// Destructor()
////////////////////////////////////////////////////////////////////////////
clMerchantableStemWoodVolume::~clMerchantableStemWoodVolume()
{
  delete[] mp_fB1;
  delete[] mp_fB2;
  delete[] mp_fSpgr;
  delete[] mp_iVolumeCodes;
  delete[] mp_iBiomassCodes;
  delete[] mp_iIndexes;
  delete[] m_cQuery;
}

/////////////////////////////////////////////////////////////////////////////
// GetData()
/////////////////////////////////////////////////////////////////////////////
void clMerchantableStemWoodVolume::GetData( xercesc::DOMDocument * p_oDoc )
{
  floatVal * p_fTempValues = NULL; //for getting species-specific values
  char * cQueryTemp = NULL;
  try
  {
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
    char cQueryPiece[5]; //for assembling the search query
    float fConvertMgToLb = CONVERT_MG_TO_KG * CONVERT_KG_TO_LB;
    int i;

    cQueryTemp = new char[(p_oPop->GetNumberOfSpecies() * 4) + 50];

    //*************************
    // Read in parameters
    //*************************

    //Set up our arrays that will extract values only for the species
    //assigned to this behavior
    p_fTempValues = new floatVal[m_iNumBehaviorSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      p_fTempValues[i].code = mp_iWhatSpecies[i];

    //Declare our arrays
    mp_fB1 = new float[m_iNumBehaviorSpecies];
    mp_fB2 = new float[m_iNumBehaviorSpecies];
    mp_fSpgr = new float[m_iNumBehaviorSpecies];

    //Get the parameter file values

    //B1 in the stem fraction equation
    FillSpeciesSpecificValue( p_oElement, "an_fiaVolB1", "an_fvb1Val",
        p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true );
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fB1[i] = p_fTempValues[i].val;

    //B2 in the stem fraction equation
    FillSpeciesSpecificValue( p_oElement, "an_fiaVolB2", "an_fvb2Val",
        p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true );
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fB2[i] = p_fTempValues[i].val;

    //SPGR in the volume equation - add in the conversion factor from Mg to lb
    FillSpeciesSpecificValue( p_oElement, "an_fiaVolSpgr", "an_fvsVal",
        p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true );
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fSpgr[i] = fConvertMgToLb / p_fTempValues[i].val;

    //*************************
    // Format query string
    //*************************

    //Do a type/species search on all the types and species
    strcpy( cQueryTemp, "species=" );
    for ( i = 0; i < m_iNumBehaviorSpecies - 1; i++ )
    {
      sprintf( cQueryPiece, "%d%s", mp_iWhatSpecies[i], "," );
      strcat( cQueryTemp, cQueryPiece );
    }
    sprintf( cQueryPiece, "%d", mp_iWhatSpecies[m_iNumBehaviorSpecies - 1] );
    strcat( cQueryTemp, cQueryPiece );
    sprintf( cQueryPiece, "%s%d", "::type=", clTreePopulation::adult);
    strcat( cQueryTemp, cQueryPiece );

    //Now put it in m_cQuery, sized correctly
    m_cQuery = new char[strlen( cQueryTemp ) + 1];
    strcpy( m_cQuery, cQueryTemp );

    //*************************
    // Get biomass codes
    //*************************
    mp_iBiomassCodes = new short int[m_iNumBehaviorSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ ) {
      mp_iBiomassCodes[i] = -1;
      mp_iBiomassCodes[i] = p_oPop->GetFloatDataCode("Biomass",
          mp_iWhatSpecies[i], clTreePopulation::adult);
      if (-1 == mp_iBiomassCodes[i]) {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clMerchantableStemWoodVolume::GetData" );
        strcpy( stcErr.cMoreInfo, "You must use Dimension Analysis with FIA Merchantable Stem Volume." );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }
    }

    delete[] cQueryTemp;
    delete[] p_fTempValues;

    Action();
  }
  catch ( modelErr & err )
  {
    delete[] p_fTempValues;
    delete[] cQueryTemp;
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    delete[] p_fTempValues;
    delete[] cQueryTemp;
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clMerchantableStemWoodVolume::GetData" );
    throw( stcErr );
  }
}



////////////////////////////////////////////////////////////////////////////
// Action()
////////////////////////////////////////////////////////////////////////////
void clMerchantableStemWoodVolume::Action()
{
  try
  {
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreeSearch * p_oBehaviorTrees;
    clTree * p_oTree;
    float fBiomass, //tree's biomass
          fStemFrac,
          fVolume,
          fDBH; //tree's dbh
    int iSp, iTp;

    //Get the trees that apply to this behavior
    p_oBehaviorTrees = p_oPop->Find( m_cQuery );

    p_oTree = p_oBehaviorTrees->NextTree();
    while ( p_oTree )
    {

      iSp = p_oTree->GetSpecies();
      iTp = p_oTree->GetType();

      //Double-check this tree's appropriateness by making sure we have a
      //non -1 return code for the volume data member
      if ( -1 != mp_iVolumeCodes[mp_iIndexes[iSp]] )
      {

        //Get this tree's DBH & biomass
        p_oTree->GetValue( p_oPop->GetDbhCode( iSp, iTp ), & fDBH );
        p_oTree->GetValue(mp_iBiomassCodes[mp_iIndexes[iSp]], &fBiomass);

        //Get stem fraction
        fStemFrac = exp(mp_fB1[mp_iIndexes[iSp]] + mp_fB2[mp_iIndexes[iSp]]/fDBH);

        //Get volume
        fVolume = fBiomass * fStemFrac * mp_fSpgr[mp_iIndexes[iSp]];

        p_oTree->SetValue( mp_iVolumeCodes[mp_iIndexes[iSp]], fVolume);
      }

      p_oTree = p_oBehaviorTrees->NextTree();
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clMerchantableStemWoodVolume::Action" );
    throw( stcErr );
  }
}

////////////////////////////////////////////////////////////////////////////
// RegisterTreeDataMembers()
////////////////////////////////////////////////////////////////////////////
void clMerchantableStemWoodVolume::RegisterTreeDataMembers()
{

  clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
  short int iNumSpecies = p_oPop->GetNumberOfSpecies(),
            i;

  //Make the list of indexes
  mp_iIndexes = new int[iNumSpecies];
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_iIndexes[mp_iWhatSpecies[i]] = i;


  //Declare the array and register our new data member
  mp_iVolumeCodes = new short int [m_iNumBehaviorSpecies];

  for ( i = 0; i < m_iNumSpeciesTypeCombos; i++ ) {
    //Make sure that only allowed types have been applied
    if ( clTreePopulation::adult != mp_whatSpeciesTypeCombos[i].iType) {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clMerchantableStemWoodVolume::RegisterTreeDataMembers" );
      strcpy( stcErr.cMoreInfo, "This behavior can only be applied to adults." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }

    //Register the code and capture it
    mp_iVolumeCodes[mp_iIndexes[mp_whatSpeciesTypeCombos[i].iSpecies]] =
         p_oPop->RegisterFloat( "FIA_Volume", mp_whatSpeciesTypeCombos[i].iSpecies, mp_whatSpeciesTypeCombos[i].iType );
  }
}
